// -*- c++ -*-

//
//      CondProb.java (Conditional Probability of Bayesian Net)
//                     Multi Layer Neural Network is included
//
//	Programmed by Y.Motomura(motomura@etl.go.jp)
//

import java.lang.*;
import java.io.*;
import java.util.*;
import java.net.*;

class CondProb extends Thread{

    // Valiables
    int Hidden = 10;
	int insize, outsize;
	int Loop, seed, patterns;
	int MaxLoop = 500;
	Node node;
	double x[][];
	double tmp[], h[];
	double wih[][];
	double who[][];
	double tho[], th[];
    double  t[][], dwih[][], dwho[][];
    double  dth[], dtho[] ,o[];
    double  d2[], d3[];
    int maxPattern = 2000;

	double eta = 0.01;
	double alpha = 0.005;
	double ErrorFunc = 0.0005;
	double errorfunc, errorfunc2;
	double errors[];
	double Wmin = -0.5;
	double Wmax = 0.5;
//	double interval = 25;
	static String DataDirectory = "Net/";
	static String WeightFile;
	static String PatternFile;

	DataBase DB;
	String DataSetName = new String();
	DataBaseFrame DBF;
	CondProbFrame CPF;
	DataFrame DF;
	int normalize, linear;
	int npatterns;
	boolean NormalizedData = false;
	double PatternMax,PatternMin;
// Constracter

  CondProb(Node n, int ins, int outs) {
      this(n, ins, outs, 10);
  }

  CondProb(Node n, int ins, int outs, int hids) {
	node = n; insize = ins; outsize = outs; Hidden = hids;
	initParameters();
	initFile();

//  DataBase Access
    DB = new DataBase();         // Databese Object
 	CPF = new CondProbFrame(this);  // ControlPanel of this class
 	DF = new DataFrame();           // Show sample data
	CPF.show();
  }
 // Initializers
  public void initParameters(){
    wih = new double[insize][Hidden];
	who = new double[Hidden][outsize];
	tho = new double[outsize]; th = new double[Hidden];
    tmp = new double[outsize];
	h = new double[Hidden];
	dwih = new double[insize][Hidden];
	dwho = new double[Hidden][outsize];
	dth = new double[Hidden]; dtho = new double[outsize];
	o = new double[outsize];
	d2 = new double[Hidden]; d3 = new double[outsize];
	errorfunc = 0;
	initWeights();
  }

  void initWeights(){
    int i,k;
    Loop=0;
    for( i = 0; i < outsize ; i++ ){ tho[i] = 0.0; }
    for( k = 0; k < Hidden; k++ ){
      for( i = 0; i < insize; i++ ) { wih[i][k] = Random(0, Wmax); }
      for( i = 0; i < outsize ; i++ ){ who[k][i] = Random(); }
      th[k] = Random() ;
    }
  }

  public void initFile () {
 	WeightFile = new String(DataDirectory + node.name + ".net");
//	PatternFile = new String(DataDirectory + node.name + ".pat");
  }

   public void CreateDBFrame(){
    if (DB == null){ DB = new DataBase(); }
    if (DBF == null){ DBF = new DataBaseFrame(DB); }
    DBF.show();
  }

  public void setDataSetName (String N) {
	DataSetName = N;
  }

  public void setLearingParameters(double e, double a){
    eta=e; alpha=a;
  }

  public void ChDataDir (String str) {
	DataDirectory = str;
	initFile();
  }
  // Functions
  double Random(){
    return Random(Wmax, Wmin);
  }
  double Random(double min, double max){
    double a = Math.random() * (max - min) + min;
   // System.out.println(Double.toString(a));
    return a;
  }

  double Sigmoid(double x){
	return((double)1.0 / (1.0 + Math.exp( -1.0*(x))));
  }

 // Pro[agation Network
  public void getOut(){
	int i,j,a;

      j = 0;
      for (Enumeration e = node.inNodes.elements(); e.hasMoreElements();) {
     	Node n = (Node)e.nextElement();
	    a = Math.abs(n.Type);
	    for(i=0; i<a; i++){
	      x[0][j] = n.outval[i]; j++;
	    }
      }
      if(node.Type>0){calcDiscrete();}
      else{calcContinuous();}
  }
  public void setOutVal(){
   int i;
   for(i=0; i< outsize;i++){
	node.outval[i] = o[i];
    }
  }
// Propagation for Neural Network
  void calcContinuous(){    Propagation(0,0);  }
  void calcDiscrete(){    Propagation(0,1);  }

  void Propagate(int mode){
     Propagate(npatterns,mode);
  }

  void Propagate(int p,int mode){
    int   i, j;
    double net;

    for( i = 0; i < Hidden; i++ ) {
      net = 0.0;
      for(j = 0; j < insize; j++){
           net += wih[j][i] * x[p][j];
      }
      h[i] = Sigmoid( net + th[i] );
    }

    for( i = 0; i < outsize; i++ ) {
      net = 0.0;
      for(j = 0; j < Hidden; j++){
        net += who[j][i] * h[j];
      }
      if (mode == 0 ){
          tmp[i] = net + tho[i];
      }else{
           tmp[i] = Sigmoid(net + tho[i]);
      }
    }
  }

 void Propagation (int p, int mode){
    int   i, j;
    double net, z1, z2;
      if(mode == 0){
         Propagate(p, 0 );
         z2 = 1.0;
     }
     else{
        Propagate(p, 1);
        for(z1 = 0.0, i=0; i< outsize;i++){
           z1 += tmp[i];
        }
        for(z2 = 0.0, i=0; i< outsize;i++){
           tmp[i] = tmp[i] / z1;
           if(tmp[i]>0.95) tmp[i] = 1.0;
           if(tmp[i]<0.05) tmp[i] = 0.0;
           z2+=tmp[i];
        }

      }
      for(i=0; i< outsize;i++){
	      o[i] = tmp[i]/z2;  // outputs
      }
 }

  void  BackPropagation( int p, int mode )
 {
        int     i, j;
        double  sum;

  for( i = 0; i < outsize; i++ ){

   if (mode == 0){
 //   d3[i] = Sigmoid( t[p][i] - o[i] ) - 0.5;
    d3[i] = t[p][i] - o[i];
   } else {
    d3[i] = ( t[p][i] - o[i] ) * o[i] * ( 1 - o[i] ) ;
   }
  }

  //System.out.println("d3="+Double.toString(d3[0])+" o="+Double.toString(o[0])+" t="+Double.toString(t[p][0])+" x[0]="+Double.toString(x[p][0]) +" x[1]="+Double.toString(x[p][1])   );

  for( i = 0 ; i < Hidden; i++ ) {
    for( sum = 0, j = 0; j < outsize; j++) {

      dwho[i][j] = eta * d3[j] * h[i] + alpha * dwho[i][j];
      who[i][j] += dwho[i][j];
      sum += d3[j] * who[i][j];
    }
      d2[i] = h[i] * ( 1 - h[i] ) * sum;
  }
  for( i = 0; i < outsize; i++ ) {
    dtho[i] = eta * d3[i] + alpha * dtho[i];
    tho[i] += dtho[i];
  }

  for( i = 0; i < insize; i++ )
    for( j = 0; j < Hidden; j++) {
      dwih[i][j] = eta * d2[j] * x[p][i]+alpha * dwih[i][j];
      wih[i][j] += dwih[i][j];
    }
  for( i = 0; i < Hidden; i++ ) {
    dth[i] = eta * d2[i] + alpha * dth[i];
    th[i] += dth[i];
  }
 }
/*
  public void LearnMonitor(int p, int  normalize) {
   int i;
   double osum, osum2;

    Propagation(p);
    if ( normalize == 1){
  	for(osum = 0.0, i=0; i< outsize;i++){
    	osum+=o[i];
  	}
  	for(osum2 = 0.0, i=0; i< outsize;i++){
    	o[i] = o[i] / osum;
    	if(o[i]>0.95) o[i] = 1.0;
    	if(o[i]<0.05) o[i] = 0.0;
    	osum2+=o[i];
  	}
   }else{
  	osum=osum2=1.0;
  }
  for(i=0; i< insize;i++){ System.out.println(Double.toString(x[p][i])); }
  System.out.println("\n | ");
  for(i=0; i< outsize;i++) {
	System.out.println(Double.toString(Math.abs((o[i]/osum2)-t[p][i]))); }
  System.out.println("\n | ");
  for(i=0; i< outsize;i++){ System.out.println(Double.toString(o[i]/osum2)); }
  System.out.println("\n | ");
  }
  */
  public void run(){
    int i,n ;
    for(n = 100,i=0 ; n > 1 ;i++){
       n = LearnProb(10000);
    }
  }
//
//  Discrete   : node.Type < 0 -> normalize = 1, linear =0
//  Continuous : node.Type >= 0 -> normalize =0, linear =1
//
  public int LearnProb(int maxloop) {
	if ( node.Type < 0) { return LearnProb(maxloop, 1, 0); }
	else { return LearnProb(maxloop, 0, 1); }
  }

  public int LearnProb(int maxloop, int  normalize, int linear) {
	int i,j,k;
	double osum,osum2;

        errorfunc = ErrorFunc + 1.0;
        for( i = 0; (errorfunc > ErrorFunc)&&(i<maxloop); i++ ) {
            Loop++;
          for( j = 1; j < npatterns +1 ; j++ ) {
            Propagation( j , linear);
            BackPropagation( j , linear);
          }

          errorfunc = (double)0.0;
          errorfunc2= (double)0.0;

          for( j = 1; j < npatterns +1 ; j++ ) {
            errors[j]=0;
            Propagation( j, linear );

	        if( normalize == 1){

              for(osum = 0,k = 0; k < outsize; k++){
                osum += o[k];
              }

              /*
              for(osum2 = 0.0, k=0; k< outsize;k++){
                o[k] = o[k] / osum;
                if(o[k]/osum>0.95) o[k] = 1.0;
                if(o[k]/osum<0.05) o[k] = 0.0;
                osum2+=o[k];
              }
              */
	        }else {
              osum = osum2 = 1.0 ;
	        }
	       osum = 1.0;

           for(k = 0; k < outsize; k++){
              errors[j] += Math.abs((o[k] - t[j][k])/(double)outsize);
           }
//           System.out.println("Pattern=" + Integer.toString(j)+", error="+Double.toString(errors[j]));
  //         System.out.println("errorfunc="+Double.toString(errorfunc));
           errorfunc = Math.max(errorfunc , errors[j]);
           errorfunc2 = errorfunc2+ errors[j];
         }
        //  errorfunc = errorfunc/((double) npatterns);
          (CPF.ErrorVal).setText(Double.toString(errorfunc));
          (CPF.ErrorAve).setText(Double.toString(errorfunc2/npatterns));
          (CPF.labelCount).setText(Integer.toString(Loop));
        }
        return i;
  }

  public void LoadPatterns() {
	LoadPatterns(2000);
  }

  public void LoadPatterns(int MaxPattern) {
	   InputStream is;
           StreamTokenizer st;

           int tType, i, j;

           try {
	        is = new FileInputStream(PatternFile);
            	st = new StreamTokenizer(is);
            	st.slashSlashComments(true);

            	st.eolIsSignificant(false);
		tType = st.nextToken();
            	for (patterns=0; (tType != StreamTokenizer.TT_EOF)&&(patterns < MaxPattern) ; patterns++) {
			for(j=0; j < insize; j++){
				x[patterns][j]=(double)st.nval;
				tType = st.nextToken();
			}

			for(j=0; j < outsize; j++){
				wih[j][patterns]=(double)st.nval;
				tType = st.nextToken();
			}
		}
		is.close();
	    } catch (IOException e) {
        	e.printStackTrace();
            }
            return;
	}

  public void GetPatterns() {
       int  i, j, n;
	   Vector result[] = new Vector[insize + outsize];
	   Vector result1[] = new Vector[insize + outsize];
	   double val;
	   n = insize+outsize;
       for(i = 0; i< n ; i++){
		    result[i] = new Vector();
	    }
	     for(i = 0; i< n ; i++){
		    result1[i] = new Vector();
	    }
	//   DB.connect();

    //   DB.getData(DataSetName, result);
       DBF.setData(result,0);
       DBF.setData(result1,1);
  //     DF.setData(result, n);
	   npatterns = result[0].size()+result1[0].size();
	    DF.setData(result, npatterns);
        x = new double[npatterns+ 1][insize];
	    t = new double[npatterns+ 1][outsize];
	    errors = new double[npatterns + 1];

        for (patterns=1; patterns < result[0].size() + 1  ; patterns++) {
		  for(j=0; j < insize; j++){
		    try {
		           String VAL = (String)result[j + outsize ].elementAt(patterns - 1 );
			       x[patterns][j] = (Double.valueOf(VAL)).doubleValue();
		    }catch(NumberFormatException e){
		         System.out.println("Error: String to Double");
		    }
		  }

		  for(j=0; j < outsize; j++){
		    try {
		        String VAL = (String)result[j].elementAt(patterns - 1);
			    t[patterns][j]= (Double.valueOf(VAL)).doubleValue();
			}catch(NumberFormatException e){
		         System.out.println("Error: String to Double");
		    }
		  }
	    }
	     for (patterns=1; patterns < result1[0].size() + 1  ; patterns++) {
		  for(j=0; j < insize; j++){
		    try {
		           String VAL = (String)result1[j + outsize ].elementAt(patterns - 1 );
			       x[patterns+result[0].size()][j] = (Double.valueOf(VAL)).doubleValue();
		    }catch(NumberFormatException e){
		         System.out.println("Error: String to Double");
		    }
		  }

		  for(j=0; j < outsize; j++){
		    try {
		        String VAL = (String)result1[j].elementAt(patterns - 1);
			    t[patterns][j+result[0].size()]= (Double.valueOf(VAL)).doubleValue();
			}catch(NumberFormatException e){
		         System.out.println("Error: String to Double");
		    }
		  }
	    }
	 
	 if (NormalizedData){
             normalizeData();
     }
     System.out.println("Max=" + Double.toString(Max()));
     return;
	}

	public void GetPatterns(String InTbl, String OutTbl, String DName, int  delay) {
       int  i, j, n;
	   Vector result[] = new Vector[insize + outsize];
	   double val;
	   n = insize+outsize;
       for(i = 0; i< n ; i++){
		    result[i] = new Vector();
	    }
	//   DB.connect();
	    if (DB.conn == null){ DB.connect("rwcpc8", "sysman", "sysman");}
	    DB.setTables(OutTbl);
	    DB.appendTable(InTbl);

        DB.getData(DName, result,0);
    //   DB.getData(DataSetName, result);

	   npatterns = result[0].size();
        x = new double[npatterns + 1][insize];
	    t = new double[npatterns + 1][outsize];
	    errors = new double[npatterns + 1];

        for (patterns=1; patterns < npatterns + 1  ; patterns++) {
		  for(j=0; j < insize; j++){
		    try {
		           String VAL = (String)result[j + outsize ].elementAt(patterns - 1 );
			       x[patterns][j] = (Double.valueOf(VAL)).doubleValue();
		    }catch(NumberFormatException e){
		         System.out.println("Error: String to Double");
		    }
		}

		for(j=0; j < outsize; j++){
		    try {
		        String VAL = (String)result[j].elementAt(patterns - 1);
			    t[patterns][j]= (Double.valueOf(VAL)).doubleValue();
			}catch(NumberFormatException e){
		         System.out.println("Error: String to Double");
		    }
		}
	 }
	 if (NormalizedData){
             normalizeData();
     }
     return;
	}

// Normalizers
  public void normalizeData() {
       int  i, j, p;
       double max;

       max = Max();
    System.out.println("Max="+Double.toString(max)+", "+Integer.toString(patterns));
        for (p=1; p < npatterns +1; p++) {
		  for(j=0; j < insize; j++){

			 x[p][j] = x[p][j]/max;
			// System.out.println(Double.toString(x[p][j]));
		}

		for(j=0; j < outsize; j++){
			 t[p][j]= t[p][j]/max;
			// System.out.println(Double.toString([p][j]));
		}
	 }
            return;
	}

  double Max(){
    double mmax = -1;
    int p, i;

    for (p=1; p < npatterns +1; p++) {
        for(i = 0; i< insize;i++){
           mmax = Math.max(x[p][i], mmax);
        }
        for(i = 0; i< outsize;i++){
           mmax = Math.max(t[p][i], mmax);
        }
    }
    PatternMax = mmax;
    return mmax;
  }
  double Min(){
    double mmin = 10000000;
    int p, i;

    for (p=1; p < npatterns +1; p++) {
        for(i = 0; i< insize;i++){
           mmin = Math.min(x[p][i], mmin);
        }
        for(i = 0; i< outsize;i++){
           mmin = Math.min(t[p][i], mmin);
        }
    }
    PatternMin = mmin;
    return mmin;
  }

  public void LoadWeights() {
	   InputStream is;
           StreamTokenizer st;

           int tType, i, j;

           try {
	        is = new FileInputStream(WeightFile);
//                is = new URL(BN.getDocumentBase(), fname).openStream();
            	st = new StreamTokenizer(is);
/* initialization of StreamTokenizer */
            	st.slashSlashComments(true);
//            	st.eolIsSignificant(true);

            	st.eolIsSignificant(false);
		tType = st.nextToken(); Loop = (int)st.nval;
		tType = st.nextToken(); errorfunc = (double)st.nval;
		        st.nextToken();
		for(j = 0; j < outsize ; j++){
                  	tho[j] = (double)st.nval;
			tType = st.nextToken();
		}
		i = 0;
            	for ( i=0; (tType != StreamTokenizer.TT_EOF)&&(i < Hidden); i++) {
			for(j=0; j < outsize; j++){
				who[i][j]=(double)st.nval;
				tType = st.nextToken();
			}

			for(j=0; j < insize; j++){
				wih[j][i]=(double)st.nval;
				tType = st.nextToken();
			}
			th[i]=(double)st.nval;
			tType = st.nextToken();
		}
		is.close();
	    } catch (IOException e) {
        	e.printStackTrace();
            }
//	    ContentsToString();
            System.out.println("Load from "+WeightFile);
            return;
	}

  public void SaveWeights() {
	   RandomAccessFile fd;

           int  i, j;
           try {
	        fd = new RandomAccessFile(WeightFile, "rw");
		fd.writeBytes(Integer.toString(Loop)+" "
			+Double.toString(errorfunc)+"\n");
    		for(j = 0; j < outsize ; j++){
		   fd.writeBytes(Double.toString(tho[j])+"\n");
		}
    		for(i = 0; i < Hidden ; i++){
      		  for(j = 0; j < outsize ; j++){
		   fd.writeBytes(Double.toString(who[i][j])+" ");
		  }
      		  for(j = 0; j < insize ; j++){
		   fd.writeBytes(Double.toString(wih[j][i])+" ");
		  }
		fd.writeBytes(Double.toString(th[i])+"\n");
		}
		fd.writeBytes("                                                                            \n");
		fd.close();
	    } catch (IOException e) {
        	e.printStackTrace();
            }
	    System.out.println("Save to "+WeightFile);
            return;
	}

	public void ContentsToString(){
		int i, j;
		StringBuffer s;
		s = new StringBuffer(4096);
		for(j = 0; j < outsize ; j++){
                  	s.append(Double.toString(tho[j])+"\n");
		}
            	for(i=0; i < Hidden; i++) {
			for(j=0; j < outsize; j++){
				s.append(Double.toString(who[i][j])+" ");
			}

			for(j=0; j < insize; j++){
				s.append(Double.toString(wih[j][i])+" ");
			}
			s.append(Double.toString(th[i])+"\n");
		}
		System.out.println(s.toString());
//	        System.out.println("-------------------------------");
  }
}



