// -*- c++ -*-

//
//      Field.java
//
//	Programmed by Jun Iio @ Real World Computing Partnership
//		      iiojun@trc.rwcp.or.jp
//
//      Modified by Y.Motomura(motomura@etl.go.jp)

import java.util.*;
import java.awt.*;
import java.io.*;
import java.net.URL;

class Field extends Panel implements Runnable {
    final static Color beige = new Color(245, 245, 220);
    final static Font font1 = new Font("Helvetica", Font.BOLD, 12);

    private final static int canvasOffsetX = 5;
    private final static int canvasOffsetY = 20;

    Bayonnet BN;

    Thread th;

    Node selectedNode;
    Node rootNode;

    Edge selectedEdge;

    Rectangle canvas;

    Point linkPoint;


    Image offScreen;
    Dimension offScreenSize;
    Graphics offGraphics;

    Field(Bayonnet e) {
        BN = e;
	linkPoint = new Point(0, 0);
    }

    public void run() {
        while (true) {
            repaint();
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                break;
            }
        }
    }


    public void makeNewNode(){
        selectedNode = new Node(25, 70);
        selectedNode.selected = true;
	selectedNode.Type = 1;
	BN.addNode(selectedNode);
    }

    public synchronized void update(Graphics g) {
        Dimension d = size();
        if ((offScreen == null)
            || (d.width != offScreenSize.width)
            || (d.height != offScreenSize.height)) {
            canvas.resize(d.width-10, d.height-55);
            offScreen = createImage(d.width, d.height);
            offScreenSize = d;
            offGraphics = offScreen.getGraphics();
            offGraphics.setFont(getFont());
        }

/* drawing background figure */
        offGraphics.setColor(Color.lightGray);
        offGraphics.fillRect(0, 0, d.width, d.height);

        offGraphics.setColor(beige);
        offGraphics.fillRect(canvas.x, canvas.y,
                             canvas.width, canvas.height);
        offGraphics.setColor(Color.lightGray);
	offGraphics.clipRect(0, 0, canvas.x + canvas.width,
			           canvas.y + canvas.height);
/* drawing selectedEdge */
        if (selectedEdge != null) selectedEdge.paint(offGraphics);

/* drawing Bayesian Net */
/* drawing Edges */
        BN.edges.paint(offGraphics);
/* drawing Nodes */
        BN.nodes.paint(offGraphics);

/* drawing rootNode */
        if (rootNode != null) rootNode.paint(offGraphics);

/* drawing selectedNode */
        if (selectedNode != null) selectedNode.paint(offGraphics);

        g.drawImage(offScreen, 0, 0, null);
    }

    public void start() {
        Dimension d = size();
        if (canvas == null) 
	    canvas = new Rectangle(canvasOffsetX, canvasOffsetY,
				   d.width-10, d.height-55);
        th = new Thread(this);
        th.start();
    }

    public void stop() {
        if (th != null) th.stop();
    }

    public void suspend() {
        if (th != null) th.suspend();
    }

    public void resume() {
        if (th != null) th.resume();
    }

    public synchronized boolean mouseEnter(Event evt, int x, int y) {
        if (BN.mode == BN.UNLINK) {
            return true;
        }
        return false;
    }

    public synchronized boolean mouseExit(Event evt, int x, int y) {
        if (BN.mode == BN.UNLINK) {
	    return true;
	}
	return false;
    }

    public synchronized boolean mouseMove(Event evt, int x, int y) {
	if (BN.mode == BN.UNLINK) {
	    return true;
	}
	return false;
    }

    public synchronized boolean mouseDown(Event evt, int x, int y) {
        Node node;
        Edge edge;

        // turning to MOVE mode by draging inside stackTop, 
        // whenever it is in any mode
        //
        if ((BN.mode != BN.MOVE) && selectedNode !=null ) {
	  if(selectedNode.inside(x, y)){
            selectedNode.dx = x - selectedNode.x;
            selectedNode.dy = y - selectedNode.y;
            repaint();
            BN.cp.buttonReset(BN.mode);
            BN.mode = BN.MOVE; 
            BN.cp.changeCurrent(BN.MOVE);
            return true;
	  }
        }

        switch (BN.mode) {
        case BN.MOVE:
            if (selectedNode !=null ) {
	      if(selectedNode.inside(x, y)){
                selectedNode.dx = x - selectedNode.x;
                selectedNode.dy = y - selectedNode.y;
                repaint();
                return true;
	      }
            }
            node = BN.nodes.findNodeLast(x, y);
            if (node != null) {
                selectedNode = node;
                selectedNode.selected = true;
                selectedNode.dx = x - selectedNode.x;
                selectedNode.dy = y - selectedNode.y;
                BN.nodes.removeElement(node);
                repaint();
                return true;
            }
	    break;

        case BN.LINK:
	    linkPoint.move(x, y);
            node = BN.nodes.findNodeLast(x, y);
            if (node == null) {
		return false;
	    }
            selectedEdge = new Edge(node.center(), new Point(x, y), true);
            selectedEdge.selected = true;
            selectedEdge.fromNode = node;
            repaint();
            return true;

        case BN.UNLINK:
            repaint();
            node = BN.nodes.findNodeLast(x, y);
            if (node != null) return false;
            edge = BN.edges.findEdgeLast(x, y);
            if (edge == null) {
		return false;
	    }
            BN.edges.removeElement(edge);

/* delete the link between fromNode and toNode */
            edge.fromNode.out.removeElement(edge);
            edge.toNode.in.removeElement(edge);

/* add "toNode" of deleted link to GC work area,
    if toNode is still untraced, while GC tracing */
            repaint();
            return true;
        }
        return false;
    }

    public synchronized boolean mouseDrag(Event evt, int x, int y) {
        Edge edge;
        switch (BN.mode) {
        case BN.MOVE:
            if (selectedNode != null) {
                selectedNode.x = x - selectedNode.dx;
                selectedNode.y = y - selectedNode.dy;
                selectedNode.moveEdges();
                repaint();
                return true;
            }
	    break;

        case BN.LINK:
            if (selectedEdge != null) {
                selectedEdge.to.x = linkPoint.x = x;
                selectedEdge.to.y = linkPoint.y = y;
                selectedEdge.reshape();
                repaint();
                return true;
            }
	    break;

        case BN.UNLINK:
            return true;
        }
        return false;
    }

    public synchronized boolean mouseUp(Event evt, int x, int y) {
        Node node;
        Edge edge;
        switch (BN.mode) {
        case BN.MOVE:
            if (selectedNode != null) {
                selectedNode.x = x - selectedNode.dx;
                selectedNode.y = y - selectedNode.dy;
                selectedNode.selected = false;
                if (canvas.inside(selectedNode.x, selectedNode.y) &&
		    canvas.inside(selectedNode.x + Node.defaultW - 1,
				  selectedNode.y + Node.defaultH - 1)) {
                    BN.nodes.addElement(selectedNode);
		}
		selectedNode = null; 
                repaint();
                return true;
            }
	    break;

        case BN.LINK:
            if (selectedEdge != null) {
                selectedEdge.selected = false;
            }
            node = BN.nodes.findNodeLast(linkPoint.x, linkPoint.y);
            if ((node != null) && (node != selectedEdge.fromNode)) {
                selectedEdge.toNode = node;
                selectedEdge.to = node.center();
                selectedEdge.reshape();
                BN.edges.addElement(selectedEdge);

/* make sure the link between fromNode and toNode */
                selectedEdge.fromNode.out.addElement(selectedEdge);
                node.in.addElement(selectedEdge);
            }
            repaint();
            selectedEdge = null;
            return true;

        }
        return false;
    }

    void clear() {
	selectedNode = rootNode = null;
	selectedEdge = null;
	BN.nodes.removeAllElements();
	BN.edges.removeAllElements();
	return;
    }

/* load & save method is for debugging ... */
    void load() {
	load(BN.graphFile);
    }

    void save() {
	BN.save();
    }

    private void load(String fname) {
	InputStream is;
        StreamTokenizer st;

        try {
	    is = new URL(BN.getDocumentBase(), fname).openStream();

/* initialization of StreamTokenizer */
	    st = new StreamTokenizer(is);
	    st.slashSlashComments(true);
	    st.eolIsSignificant(false);

	    BN.nodes.removeAllElements();
            BN.edges.removeAllElements();

            Node node;
            Edge edge;
	    int tType;

/* Nodes information loading ... */
            int maxId = 0;
	    while ((tType = st.nextToken()) != StreamTokenizer.TT_EOF) {
            	if ((tType == StreamTokenizer.TT_WORD) &&
                     (st.sval.equals("Edges"))) break;
                if (tType != StreamTokenizer.TT_NUMBER) continue;
// Node generateing here.
                node = new Node();
                node.nodeId = (int)st.nval;
                maxId = Math.max(maxId, node.nodeId);	
                st.nextToken();	
                node.x = (int)st.nval;
                st.nextToken();	
                node.y = (int)st.nval;
                st.nextToken();	
		node.Type = (int)st.nval; node.root = false;
		st.nextToken();
		node.name = st.sval;
		node.init();
                BN.nodes.addElement(node);
	    }
            maxId++;
            Node.idCtr = maxId;

/* Edges information loading ... */
            maxId = 0;
	    while ((tType = st.nextToken()) != StreamTokenizer.TT_EOF) {
                if (tType != StreamTokenizer.TT_NUMBER) continue;
                edge = new Edge();
                edge.edgeId = (int)st.nval;
                maxId = Math.max(maxId, edge.edgeId);	
                st.nextToken();	
                edge.from.x = (int)st.nval;
                st.nextToken();	
                edge.from.y = (int)st.nval;
                st.nextToken();	
                edge.to.x = (int)st.nval;
                st.nextToken();	
                edge.to.y = (int)st.nval;
                st.nextToken();	
                edge.fromNode = BN.nodes.pickUpById((int)st.nval);
                st.nextToken();	
                edge.toNode = BN.nodes.pickUpById((int)st.nval);
                edge.reshape();
                BN.edges.addElement(edge);
                edge.fromNode.out.addElement(edge);
                edge.toNode.in.addElement(edge);
                edge.toNode.inNodes.addElement(edge.fromNode);
		edge.toNode.child_f = true;
	    }
            maxId++;
            Edge.idCtr = maxId;
	
	    is.close();
        } catch (IOException e) {
	    e.printStackTrace();
	}
	System.out.println("Bayonnet: data loaded");
	BN.nodes.setCondProb(BN);
	System.out.println("Bayonnet: CondProbs loaded");
	return;
    }
}










