// -*- c++ -*-

//
//      FileIO.java
//
//	Programmed by Jun Iio @ Real World Computing Partnership
//		      iiojun@trc.rwcp.or.jp
//
//      Modified by Y.Motomura(motomura@etl.go.jp)

import java.util.*;
import java.awt.*;
import java.io.*;

class FileIO {
    File file;
    DataInputStream dis;
    DataOutputStream dos;
    Bayonnet BN;

    FileIO(Bayonnet b) {
	BN = b;
	file = new File(BN.graphFile);
    }

    void save(Nodes ns, Edges es) {
	FileOutputStream fos;
        try {
	    fos = new FileOutputStream(file);
	    dos = new DataOutputStream(fos);
	    dos.writeBytes("Nodes\n");
	    for (Enumeration e = ns.elements(); e.hasMoreElements();) {
		Node node = (Node)e.nextElement();
		dos.writeBytes(node.contentsToString(1));
	    }
	    dos.writeBytes("Edges\n");
	    for (Enumeration e = es.elements(); e.hasMoreElements();) {
		Edge edge = (Edge)e.nextElement();
		dos.writeBytes(edge.contentsToString());
	    }
	    dos.close();
	    fos.close();
        } catch (IOException e) {
	    e.printStackTrace();
	}
	System.out.println("GCsimulator: data saved");
    }

    void load(String file) {
        FileInputStream fis;
        StreamTokenizer st;	
        try {
	    fis = new FileInputStream(file);

/* initialization of StreamTokenizer */
	    st = new StreamTokenizer(fis);
	    st.slashSlashComments(true);
	    st.eolIsSignificant(false);

	    BN.nodes.removeAllElements();
            BN.edges.removeAllElements();

            Node node;
            Edge edge;
	    int tType;

/* Nodes information loading ... */
            int maxId = 0;
	    while ((tType = st.nextToken()) != StreamTokenizer.TT_EOF) {
            	if ((tType == StreamTokenizer.TT_WORD) &&
                     (st.sval.equals("Edges"))) break;
                if (tType != StreamTokenizer.TT_NUMBER) continue;
                node = new Node();
                node.nodeId = (int)st.nval;
                maxId = Math.max(maxId, node.nodeId);	
                st.nextToken();	
                node.x = (int)st.nval;
                st.nextToken();	
                node.y = (int)st.nval;
                st.nextToken();	
//                if (st.sval.equals("t")) { 
//                    node.root = true;
//                    BN.rootNode = node;
//                } else node.root = false;
		node.Type = (int)st.nval;
		st.nextToken();
		node.name = st.sval;
		node.init();
                BN.nodes.addElement(node);
	    }
            maxId++;
            Node.idCtr = maxId;

/* Edges information loading ... */
            maxId = 0;
	    while ((tType = st.nextToken()) != StreamTokenizer.TT_EOF) {
                if (tType != StreamTokenizer.TT_NUMBER) continue;
                edge = new Edge();
                edge.edgeId = (int)st.nval;
                maxId = Math.max(maxId, edge.edgeId);	
                st.nextToken();	
                edge.from.x = (int)st.nval;
                st.nextToken();	
                edge.from.y = (int)st.nval;
                st.nextToken();	
                edge.to.x = (int)st.nval;
                st.nextToken();	
                edge.to.y = (int)st.nval;
                st.nextToken();	
                edge.fromNode = BN.nodes.pickUpById((int)st.nval);
                st.nextToken();	
                edge.toNode = BN.nodes.pickUpById((int)st.nval);
                edge.reshape();
                BN.edges.addElement(edge);
                edge.fromNode.out.addElement(edge);
                edge.toNode.in.addElement(edge);
                edge.toNode.inNodes.addElement(edge.fromNode);
                edge.toNode.child_f = true;
	    }
            maxId++;
            Edge.idCtr = maxId;
	
	    fis.close();
        } catch (IOException e) {
	    e.printStackTrace();
	}
	System.out.println("Bayonnet: data loaded");
	BN.nodes.setCondProb(BN);
	System.out.println("Bayonnet: CondProbs loaded");
    }
}




