//
//
//  Node of Bayasian Network
//
//  Programmed by Y.Motomura(motomura@etl.go.jp)
//
import java.util.*;

import java.awt.*;

public class Node extends gNode implements TCPserver{
/* Class variable */
    public static int idCtr = 0;
    public final static int Default_PORT=6700; /* Base TCP/IP port No.(+NodeId) */
    public static boolean Node_TCP=false;     /* Listen for Each Node or not.*/
/* Status of Node */
    public boolean child_f = false;     /*  This node has parents or not*/
    public boolean evidence_f = false;  /*  This node is an evidence or not */

    public double outval[];  /* Output value of forward probability */
    public CondProb  cp;     /* Conditional Probability class(NN)   */
    Thread th;               /* for Montecarlo simulation           */
    Nodes  inNodes = null;   /* Parent nodes connected to input     */

    String TableName;
/* Followings are defined in super class(gNode) 
    int nodeId;
    String name;

    boolean selected = false;
    boolean root = false;
    boolean marked = false;
    public int Type = 0;

    Edges in, out;
*/

    Node() {
        this(defaultX, defaultY, defaultW, defaultH);
    }

    Node(int x, int y) {
	this(x, y, defaultW, defaultH);
    }

    Node(int x, int y, int w, int h) {
	super(x,y,w,h);
        nodeId = idCtr++;
        inNodes = new Nodes();
	name = new String("Node"+String.valueOf(nodeId));
	Type = 1;
    }

// get size of own output vector, and generate it
    public void init(int n){    
	Type = n;
	init();
   }
    public void init(String nm){    
	name = nm;
	init();
   }
    public void init(String nm, int n){    
	name = nm;
	Type = n;
	init();
   }
    public void init(){    
	outval = new double[Math.abs(Type)];
	if(Node_TCP){
		new TCP(nodeId+Default_PORT,this);
	}
    }

    public void getVal(){

//Now, implemented only for Feed forward Node.
	evidence_f = false;
	root = false;
	if(child_f){
	  cp.getOut();
	}else{
//Here, setting prior is required, but do nothing yet.
	}
	return;
    }

// Array version
    public void putVal(double d[]){
	int i,j;
	evidence_f = true;
	root = true;
	j = Math.abs(Type);
	for(i=0; i< j ; i++){
		outval[i] = d[i];
	}
	return;
    }

// Real value version
    public void putVal(int i, double x){

	evidence_f = true;
	root = true;
	outval[i] = x;
	return;
    }

    void setCondProb(){

// get total size of parents vector.
	int insize = 0;
	if(inNodes == null) return;
        for (Enumeration e = inNodes.elements(); e.hasMoreElements();) {
       	    Node node = (Node)e.nextElement();
	    insize += Math.abs(node.Type);
	}
	if(insize == 0) return;
// make Conditional Probability(NN) by (this, InputSize, OutputSize)
	cp = new CondProb(this, insize, Math.abs(Type));
    }

//  Load weight from file(name.net)
    void loadCondProb(){
	    cp.LoadWeights();
    }

//  Save weight to file(name.net)
    void saveCondProb(){
//	System.out.println("saveCondProb" + name );
	    cp.SaveWeights();
    }

//  Learning prosedure for condition probabilities
    public int LearnProb(int maxloop){
	    return cp.LearnProb(maxloop);
    }

//  Load pattrens for learning  from file(name.pat)
    void LoadPatterns(int maxpattern){
	// cp.LoadPatterns(maxpattern);
    }

    void GetPatterns(String DName, int delay){
        String InTable = new String();
        if(inNodes == null) return;
        for (Enumeration e = inNodes.elements(); e.hasMoreElements();) {
       	    Node node = (Node)e.nextElement();
	    InTable = InTable+" "+node.TableName;
	    }
	    System.out.println(InTable);
	   cp.GetPatterns(InTable,TableName,DName, delay);
    }
    
    String contentsToString(int pass) {
	String retStr;
        switch (pass) {
        case 1:
            retStr = Integer.toString(nodeId) + " " + 
                     Integer.toString(x) + " " +
                     Integer.toString(y) + " " +
                     Integer.toString(Type) + " \"\"\n" ;
            break;
        case 2:
            retStr = Integer.toString(nodeId) + "\nIn:";	
	    for (Enumeration e = in.elements(); e.hasMoreElements();) {
                    Edge edge = (Edge)e.nextElement();
                    retStr += " " + Integer.toString(edge.edgeId);
                }
            retStr += "\nOut:";
	    for (Enumeration e = out.elements(); e.hasMoreElements();) {
                    Edge edge = (Edge)e.nextElement();
                    retStr += " " + Integer.toString(edge.edgeId);
                }
            retStr += "\n";
            break;
        default:
            retStr = "";
        }
	return retStr;
    }

    void paint(Graphics g) {
	super.paint(g);
    }

// real process method service(input, output) used in TCP.class is defined here.
	public void service(String in, StringBuffer out){
		int i;
		StringTokenizer st;
		String word;

System.out.println("MSG@"+ name + "=" + in);
		st = new StringTokenizer(in, "()\"\n\t\r ");
		if(st.hasMoreTokens()!=true) return;
		word = new String(st.nextToken());
		if("set".equals(word)){ 
		    for(i=0; (i<outval.length)&&(st.hasMoreTokens()); i++){
			outval[i]=new Double(st.nextToken("( ),\n\t\r\"")).doubleValue();
		    }
		}else if("get".equals(word)){ 
		    if(st.hasMoreTokens()){ 
			word = new String(st.nextToken()); 
		    	if("backward".equals(word)){
// Not yet			getVal2();
		    	}else{
		    	     if("forward".equals(word)){ getVal(); }
			    }
		    }
		    out.append("( ");
		    for(i=0; i<outval.length; i++){
			    out.append(Double.toString(outval[i]));
			    out.append(" ");
		    }
		    out.append(")");
		}else if("SetTable".equals(word)){
		    TableName = new String(st.nextToken());
		}else if("CondProb".equals(word)){ 
		    if(st.hasMoreTokens()){ 
			word = new String(st.nextToken()); 
		    	if("set".equals(word)){
				setCondProb();
			}else if("Load".equals(word)){ 
				loadCondProb();
			}else if("Save".equals(word)){ 
				saveCondProb();
		//  }else if("LoadPattern".equals(word)){ 
		//		LoadPatterns(Integer.parseInt(st.nextToken()));
		    }else if("GetPatterns".equals(word)){
			    GetPatterns( st.nextToken(),Integer.parseInt(st.nextToken()));
			}else if("Learn".equals(word)){
			    LearnProb( Integer.parseInt(st.nextToken()));
			}else {
                System.out.println("MSG@"+ name + " not support");
		    	     }
			}
	    }else {
            System.out.println("MSG@"+ name + " not support");
	    }
	}
}
