// -*- c++ -*-

//
//      (graphical)Edge.java
//
//	Programmed by Jun Iio @ Real World Computing Partnership
//		      iiojun@trc.rwcp.or.jp
//
//      Modified by Y.Motomura(motomura@etl.go.jp)


import java.awt.*;

class gEdge extends Polygon {
    final static Color normalColor = Color.blue.darker();
    final static Color selectedColor = Color.cyan.darker();
    final static Color inMarkingColor = Color.green.darker();
    final static Color markedColor = Color.green.darker().darker().darker();

    static int idCtr = 0;
    int edgeId;

    boolean inmarking;
    boolean selected;
    boolean marked;
    Point from, to;
    Node fromNode, toNode;

    gEdge() {
        edgeId = idCtr++;

        from = new Point(0, 0);
        to = new Point(0, 0);

        inmarking = false;
        selected = false;
        marked = false;
        npoints = 7;
        xpoints = new int[7];
        ypoints = new int[7];
        reshape();
    }

    gEdge(Point f, Point t, boolean b) {
        edgeId = idCtr++;

        from = f;
        to = t;
        selected = b;
        inmarking = false;
        marked = false;
        npoints = 7;
        xpoints = new int[7];
        ypoints = new int[7];
        reshape();
    }

    gEdge(Node fn, Node tn) {
	this(fn.center(), tn.center(), false);
	fromNode = fn;
	toNode = tn;
    }

    void paint(Graphics g) {
        if (selected) g.setColor(selectedColor);
        else if (marked) g.setColor(markedColor);
        else if (inmarking) g.setColor(inMarkingColor);
        else g.setColor(normalColor);
        g.fillPolygon(this);
        g.setColor(Color.white);
        g.drawPolygon(this);
    }

    void reshape() {
        if (from.equals(to)) return;

//        double dw = 2.5;             /* width of arrow /2 */
        double dw = 1.5;             /* width of arrow /2 */
        int fx = from.x;
        int fy = from.y;
        int tx = to.x;
        int ty = to.y;

        double l = Math.sqrt((tx-fx)*(tx-fx)+(ty-fy)*(ty-fy));
        double dl = Math.max(l - 10*dw, 0.5*l);
        
        xpoints[0] = (int)(fx - dw*(ty-fy)/l);
        ypoints[0] = (int)(fy + dw*(tx-fx)/l);
        xpoints[1] = (int)(fx + (dl*(tx-fx) - dw*(ty-fy))/l);
        ypoints[1] = (int)(fy + (dl*(ty-fy) + dw*(tx-fx))/l);
        xpoints[2] = (int)(fx + (dl*(tx-fx) - 3*dw*(ty-fy))/l);
        ypoints[2] = (int)(fy + (dl*(ty-fy) + 3*dw*(tx-fx))/l);
        xpoints[3] = tx;
        ypoints[3] = ty;
        xpoints[4] = (int)(fx + (dl*(tx-fx) + 3*dw*(ty-fy))/l);
        ypoints[4] = (int)(fy + (dl*(ty-fy) - 3*dw*(tx-fx))/l);
        xpoints[5] = (int)(fx + (dl*(tx-fx) + dw*(ty-fy))/l);
        ypoints[5] = (int)(fy + (dl*(ty-fy) - dw*(tx-fx))/l);
        xpoints[6] = (int)(fx + dw*(ty-fy)/l);
        ypoints[6] = (int)(fy - dw*(tx-fx)/l);
    }

    String contentsToString() {
	String retStr;
        retStr = Integer.toString(edgeId) + " " + 
                 Integer.toString(from.x) + " " +
                 Integer.toString(from.y) + " " +
                 Integer.toString(to.x) + " " +
                 Integer.toString(to.y) + " " +
                 Integer.toString(fromNode.nodeId) + " " +
                 Integer.toString(toNode.nodeId) + "\n";
	return retStr;
    }
}
